#include "PictureView.h"
#include <AppKit/NSAffineTransform.h>
#include <AppKit/NSImage.h>
#include <Foundation/NSException.h>
#include <AppKit/NSEvent.h>
#include <AppKit/NSWindow.h>
#include <Foundation/NSGeometry.h>
#include <Foundation/NSData.h>
#include <AppKit/NSApplication.h>
#include <AppKit/NSMenu.h>
#include <math.h>
#include "PHImage.h"
#include "PHAffineTransform.h"
#include "Selection.h"
#include "Fragment.h"
#include "Stack.h"
#include "PHGeometry.h"

#define INTEGRAL_SIZE

@interface PictureView (PrivateMethods)

- (void) prepareCoordinates;
- (float) scaleToFitIntoSize: (NSSize) frame_size;
- (NSAffineTransform *) transformToFitIntoSize: (NSSize) frame_size;

- (void) drawRect:(NSRect) rect;

- (void) updateClipMenuItem;
- (void) updateUndoMenuItem;
    
@end // @interface PictureView (PrivateMethods)


@implementation PictureView

- (id) initWithFrame: (NSRect) rect
{
    [super initWithFrame: rect];
    _image = nil;
    _sketch = nil;

    _scale = 1.0;

    _fragm = nil;
    _fragms = [[Stack alloc] init];

    _fragm_to_sketch = nil;

    _selection = nil;

    return self;
}

- (void) dealloc
{
    RELEASE(_image);
    RELEASE(_sketch);

    RELEASE(_fragm);
    RELEASE(_fragms);
    RELEASE(_fragm_to_sketch);

    RELEASE(_selection);
}

- (void) setImage: (NSImage *) image
{
    // Keep original image
    
    ASSIGN(_image, image);

    ASSIGN(_fragm, [Fragment fragmentWithImage: image]);

    [_fragms removeAllObjects];

    DESTROY(_selection);

    [self prepareCoordinates];
    [self setNeedsDisplay: YES];

    [self updateClipMenuItem];
    [self updateUndoMenuItem];
}

- (void) clip
{
    if (_selection == nil)
        return;

    Fragment * new_fragm =
        [Fragment fragmentWithScene: [_selection selectedInImage]
                  scale: [_fragm scale]
                  rotation: [_fragm rotation]];

    [_fragms push: [_fragm restoreOriginalState]];
    
    ASSIGN(_fragm, new_fragm);

    DESTROY(_selection);

    [self prepareCoordinates];

    [self setNeedsDisplay: YES];

    [self updateClipMenuItem];
    [self updateUndoMenuItem];
}

- (void) undo
{
    if ([_fragms depth] > 0)
    {
        ASSIGN(_fragm, [_fragms pop]);

        DESTROY(_selection);

        [self prepareCoordinates];

        [self setNeedsDisplay: YES];
    }

    [self updateUndoMenuItem];
}

- (BOOL) saveSceneToFile: (NSString *) path
{
    NSImage * img = [_image imageFromRect: [_fragm scene]
                            rotatedBy: [_fragm rotation]];
    
    NSData * data = [img TIFFRepresentation];

    NSLog(@"save: [img TIFFRepresentation]=%d", data);
    
    BOOL ret = [data writeToFile: path atomically: NO];

    NSLog(@"save: [data writeToFile:]=%d", ret);

    return ret;
}

- (void) prepareCoordinates
{
    ASSIGN(_fragm_to_sketch, [self transformToFitIntoSize: [self frame].size]);

    ASSIGN(_sketch, [_image imageFromRect: [_fragm scene]
                            scaledBy: _scale
                            rotatedBy: [_fragm rotation]]);

    NSSize frame_size = [self frame].size;
    NSSize sketch_size = [_sketch size];

    NSPoint sketch_origin = NSMakePoint(
        (frame_size.width - sketch_size.width)/2,
        (frame_size.height - sketch_size.height)/2);

#ifdef INTEGRAL_SIZE
    sketch_origin.x = (int) sketch_origin.x;
    sketch_origin.y = (int) sketch_origin.y;
#endif //INTEGRAL_SIZE    
    
    [self setBoundsOrigin: NSMakePoint(-sketch_origin.x, -sketch_origin.y)];

    [_selection setFragment: [_fragm scene] xform: _fragm_to_sketch];
}

- (void) rotate
{
    [self rotateWithDirection: PHPositiveDirection];
}

- (void) rotateWithDirection: (PHRotateDirection) direction
{
    int rotation = 0;
    switch (direction)
    {
        case PHPositiveDirection: rotation = 90; break;
        case PHNegativeDirection: rotation = -90; break;
        default:
            NSLog(@"rotateWithDirection: unknown direction %d\n", direction);
            break;
    }


    [_fragm rotateByDegrees: rotation];
    [self prepareCoordinates];
    [self setNeedsDisplay: YES];
}

- (void) windowDidResize: (NSNotification *)notification
{
    [self prepareCoordinates];
    [self setNeedsDisplay: YES];
}

- (void) drawRect:(NSRect)frame
{
    if (! _sketch)
        return;

    // draw full sketch ignoring the 'frame' rect
    [_sketch compositeToPoint: NSZeroPoint operation: NSCompositeCopy];

    /*
    // draw the intersection between the sketch and the frame rectangles -
    // off-pixel problem (rounding)?
    
    NSRect sketchRect =
        NSMakeRect(0, 0, [_sketch size].width, [_sketch size].height);
    frame = NSIntersectionRect(frame, sketchRect);
        
    [_sketch compositeToPoint: frame.origin fromRect: frame
             operation: NSCompositeCopy];
    */

    if (_selection != nil)
    {
        /*
        // draw full mask ignoring the 'frame' rect
        [[_selection mask] compositeToPoint: NSZeroPoint
                           operation: NSCompositeSourceOver];
        */

        NSRect maskRect = NSInsetRect(frame, -1, -1);
        NSImage * mask = [_selection maskForRect: maskRect];

        
        [mask compositeToPoint: maskRect.origin
              operation: NSCompositeSourceOver];
    }
}

- (float) scaleToFitIntoSize: (NSSize) frame_size
{
    NSSize orig_size = [_fragm scene].size;
    
    NSSize rotated_size;
    switch ([_fragm rotation])
    {
        case 0:
        case 180:
            rotated_size = orig_size;
            break;
        case 90:
        case 270:
            rotated_size.width = orig_size.height;
            rotated_size.height = orig_size.width;
            break;
        default:
            NSAssert1(0, @"Invalid rotation value %d", [_fragm rotation]);
            break;
    }
    
    float scale_x = MIN(frame_size.width / rotated_size.width, 1);
    float scale_y = MIN(frame_size.height / rotated_size.height, 1);
    float scale = MIN(scale_x, scale_y);
    return scale;
}
    
- (NSAffineTransform *) transformToFitIntoSize: (NSSize) frame_size
{
    _scale = [self scaleToFitIntoSize: frame_size];

    return [NSAffineTransform transformToQIofRect: [_fragm scene]
                              scale: _scale
                              rotation: [_fragm rotation]];
}

- (void) mouseDown:(NSEvent *) event
{
    NSPoint pt = [self convertPoint: [event locationInWindow] fromView: nil];
    
    RELEASE(_selection);
    _selection = [[Selection alloc] initWithRect: [_fragm scene]
                                    Xform: _fragm_to_sketch
                                    Start: pt];

    [self setNeedsDisplayInRect: [_selection updateInSketch]];
    [self updateClipMenuItem];
}

- (void) mouseUp:(NSEvent *) event 
{
    if (_selection != nil)
    {
        NSPoint pt = [self convertPoint:[event locationInWindow] fromView:nil];
        
        [_selection mouseUp: pt];
        
        if ([_selection valid] == NO)
        {
            DESTROY(_selection);
            [self setNeedsDisplay: YES];
        }
        else
        {
            [self setNeedsDisplayInRect: [_selection updateInSketch]];
        }
    }
    [self updateClipMenuItem];
    
}
 
- (void) mouseDragged:(NSEvent *) event
{
    if (_selection != nil)
    {
        NSPoint pt = [self convertPoint:[event locationInWindow] fromView:nil];
        
        [_selection mouseDragged: pt];

        [self setNeedsDisplayInRect: [_selection updateInSketch]];
    }
}

- (void) updateClipMenuItem
{
    NSMenuItem * item = [[NSApp mainMenu] itemWithTitle: @"Clip"];
    
    //[item setEnabled: _selection];
    [item setAction: _selection ? @selector(clip:) : @selector(dummy:)];
}

- (void) updateUndoMenuItem
{
    NSMenuItem * item = [[NSApp mainMenu] itemWithTitle: @"Undo"];
    
    //[item  setEnabled: ([_fragms empty] == NO)];
    [item setAction:
              ([_fragms empty] == NO) ? @selector(undo:) : @selector(dummy:)];
}

@end
