#include "Selection.h"
#include <AppKit/NSAffineTransform.h>
#include <AppKit/NSEvent.h>
#include <AppKit/NSColor.h>
#include <AppKit/NSImage.h>
#include "PHGeometry.h"
#include "PHAffineTransform.h"

#define INTEGRAL_SIZE

@implementation Selection

- (id) init
{
    _fragment = NSZeroRect;
    _sketch = NSZeroRect;
    _start = NSZeroPoint;
    _end = NSZeroPoint;
    _previousSelection = NSZeroRect;

    _xform = nil;

    return self;
}

- (void) dealloc
{
    RELEASE(_xform);
}

- (id) initWithRect: (NSRect) fragment
              Xform: (NSAffineTransform *) xform
              Start: (NSPoint) start
{
    _xform = xform;
    RETAIN(_xform);

    _fragment = fragment;
    
    _sketch = [_xform transformRect: _fragment];

#ifdef INTEGRAL_SIZE
    _sketch = NSIntegralRect(_sketch);
    _sketch.origin = NSZeroPoint;
#endif //  INTEGRAL_SIZE
    
    _start = start;
    _end = _start;
    _previousSelection = _sketch; // the whole sketch

    return self;
}

- (void) setFragment: (NSRect) fragment xform: (NSAffineTransform *) xform
{
    if (_xform != xform)
    {
        _fragment = fragment;
        
        NSAffineTransform * reverse = PHReverseTransform(_xform);
        NSPoint orig_start = [reverse transformPoint: _start];
        NSPoint orig_end = [reverse transformPoint: _end];
        
        RELEASE(_xform);
        _xform = xform;
        RETAIN(xform);

        // need to convert _start, _end and _sketch
        _start = [_xform transformPoint: orig_start];
        _end = [_xform transformPoint: orig_end];
        _sketch = [_xform transformRect: _fragment];
        
#ifdef INTEGRAL_SIZE
        _sketch = NSIntegralRect(_sketch);
        _sketch.origin = NSZeroPoint;
#endif //  INTEGRAL_SIZE

    }
}

- (void) mouseDown: (NSPoint) pt
{
    _start = pt;
    _end = _start;
}

- (void) mouseDragged: (NSPoint) pt
{
    const float TOLERANCE = 5; // tolerance = 5 pixels in sketch

    _previousSelection = [self selectedInSketch];
    
    _end = (PHDistance(_start, pt) > TOLERANCE) ? pt : _start;
}

- (void) mouseUp: (NSPoint) pt
{
    [self mouseDragged: pt];
}

- (BOOL) valid
{
    return NSIsEmptyRect([self selectedInSketch]) == NO;
}


- (NSRect) userSelRect
{
    // This method should take into account the shape of the selection
    // (4:3, 1:1, arbitrary).
    // Meanwhile I implement just arbitrary.

    return PHRectFromPoints(_start, _end);
}

- (NSRect) selectedInSketch
{
    return NSIntersectionRect(_sketch, [self userSelRect]);
}

- (NSRect) updateInSketch
{
    /*
    return NSUnionRect(_previousSelection, [self selectedInSketch]);
    */

    return NSInsetRect(
        NSUnionRect(_previousSelection, [self selectedInSketch]),
        -1, -1);
}

- (NSRect) selectedInImage
{
    NSAffineTransform * reverse = PHReverseTransform(_xform);

    return NSIntersectionRect( _fragment,
                               [reverse transformRect: [self userSelRect]] );
}

- (NSImage *) mask
{
    NSImage * mask = [[NSImage alloc] initWithSize: _sketch.size];
    
    NSColor * blur = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.57];
    NSColor * transp = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.0];

    [mask lockFocus];
    
    [blur set];

    NSRectFillUsingOperation(_sketch, NSCompositeCopy);

    [transp set];

    NSRect selectionRect = [self selectedInSketch];
    
    NSRectFillUsingOperation(selectionRect, NSCompositeCopy);
    
    [mask unlockFocus];

    return AUTORELEASE(mask);
}

- (NSImage *) maskForRect: (NSRect) rect
{
    NSImage * mask = [[NSImage alloc] initWithSize: rect.size];

    NSColor * blur = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.57];
    NSColor * transp = [NSColor colorWithCalibratedWhite: 0.0 alpha: 0.0];

    [mask lockFocus];
    
    [blur set];

    NSRect maskRect = NSMakeRect(0, 0, rect.size.width, rect.size.height);

    // sketch rect in mask coordinates
    NSRect imageRect = NSOffsetRect(_sketch, -rect.origin.x, -rect.origin.y);
    
    NSRectFillUsingOperation(NSIntersectionRect(maskRect, imageRect),
                             NSCompositeCopy);

    [transp set];

    // in sketch coordinates
    NSRect selectionRect = [self selectedInSketch];

    // convert to mask coordinates
    NSRect maskSelRect = NSOffsetRect(selectionRect,
                                      -rect.origin.x, -rect.origin.y);

    maskSelRect = NSIntersectionRect(maskSelRect, maskRect);
    
    NSRectFillUsingOperation(NSIntersectionRect(maskSelRect, imageRect),
                             NSCompositeCopy);
    
    [mask unlockFocus];

    return AUTORELEASE(mask);
}

@end // implementation Selection
